%% Optimisation
clear
close all
%% Overfitting
x = [0:9];

y = 2 * x'; % data (just a straight line)
yn = y + 3*randn(10,1); % noisy data

% plot clean and noisy data
plot(x, y, 'r', x, yn, 'b')

% fit a line
% "model" is a line
Y = [0:9]';

% fit model to clean and noisy data
b = pinv(Y)*y;
b = pinv(Y)*yn;

% There are many ways to do a regression
b2 = regress(y,Y);

plot(x, y, 'r', x, yn, 'b', x, Y*b, 'm--')

%% fit data with 10 independent polynomials

% create a matrix Y with polynomials as basis functions
for i = 1:10
    Y(:,i) = (x-4.5).^i; 
end
plot(Y)

% check whether the rank is equal to number of samples
% i.e. they are linearly independent
rank(Y)

% compute the pseudoinverse
Ypinv = pinv(Y);

% estimate the parameters from clean data
b = Ypinv*y;
hold on
plot(Y*b);

% estimate parameters from noisy data
bn = Ypinv*yn;
plot(Y*bn);

% you can use your parameters to predict new data, between previous samples
a = [0:0.01:9];

data = [0];
% weight the basis functions with their estimated parameters
for i=1:10, 
    data = data + bn(i)*(a-4.5).^i; 
end

% compare fitted curve with original samples
plot(a, data)
hold on
plot(x, yn, 'rx')

% close all

%% now model the line with sines and cosines
% create a matrix with sine/cosine colums as basis functions
clear Y

% x = linspace(1,10,1000);
x = [0:9]'
for i=1:2:10, 
    Y(:,i)=sin(i*x*(pi/10)); 
    Y(:,i+1)=cos(i*x*(pi/10)); 
end

figure('Color','white')
ii = 0;
for mm = 1:10
    for nn = 1:10
        ii = ii + 1;
        subplot(10, 10, ii)
        plot(Y(:,mm), Y(:,nn))
        axis off
    end
end

%% check whether basis functions are linearly independent
close all
figure
rank(Y)
imagesc(corr(Y))

Ypinv = pinv(Y);

% estimate the parameters from clean data
figure
b = Ypinv*y;
plot(Y*b);

% estimate parameters from noisy data
figure
bn = Ypinv*yn;
plot(Y*bn);

%% you can use your parameters to predict new data, between previous samples
close all
a = [0:0.01:9];
data = [0];

figure

% weight the basis functions with their estimated parameters
for i=1:2:10, 
    s1 = bn(i)*sin(i*a*(pi/10));
    s2 = bn(i+1)*cos(i*a*(pi/10));
    data = data + s1 + s2;
    subplot(1,2,1)
    plot(a, data)
    hold on
    plot(x, yn, 'rx')
    hold off
    subplot(1,2,2)
    hold on
    plot(a, s1)
    pause(1)
end
hold on 


%% optimisation tools in Matlab

% finding minimum of Matlab's built-in functions
x = fminsearch(@cos,1)
% see also fminbnd


% finding minima for self-defined arithmetic functions
fun = @(x)(0.1*x+cos(x))
x = [-10:0.1:10];
plot(x,fun(x))
% find minima with different starting points
fminsearch(fun,1)
fminsearch(fun,-1)

% finding minima for any self-defined function
fminsearch(@my_fun,1)  % where "fun" is a function

% finding minima for a function with multiple input parameters
% note: fminbnd vs fminsearch
fminbnd(@(p) find_phase(p, 1, sin([1:100]+1)), -pi, pi)






%%----------------------------------------------------------------
%%------------------   GENETIC ALGORITHM   -----------------------
%%----------------------------------------------------------------

%% Choose an Image
clc
close all
clear all
IM = imread('/imaging/dp01/optimisation/heart.jpg');

IM = IM(:,:,1);
imagesc(IM)
colormap gray

sz = size(IM);
sz = round(sz/4);
IM = double(imresize(IM, sz));
imagesc(IM)

% Set up variables for genetic algorithm

tic

close all
C = zeros(sz);

N = 200;
clear gen_*
% Position of the dots X, Y coordinates
gen_pos(:,1) = (rand(N,1)*(sz(1)-1)) + 1;
gen_pos(:,2) = (rand(N,1)*(sz(2)-1)) + 1;
gen_pos = round(gen_pos); % round off to pixel coordinates
cvec = zeros(1,100);
% Starting point for utility function (correlation between dots and the selected image. 
% (this is what we are trying to maximise)
c_best = -1; 
gen_intensity = ones(N,1);

%% Run Algorithm
close all
gen_size = sz(1)/40;
clear rem
Nmutations = 1;
ind = 0;

F = fspecial('disk', gen_size);

while c_best < 1
    ind = ind + 1;

    % MUTATE

    alterpos_index = round(rand(Nmutations,1)*(N-1))+1;
    gen_pos_new = gen_pos;
    gen_intensity_new = gen_intensity;
    gen_pos_new(alterpos_index,1) = round(rand(Nmutations,1)*(sz(1)-1)) + 1;
    gen_pos_new(alterpos_index,2) = round(rand(Nmutations,1)*(sz(2)-1)) + 1;
    
    RGB = zeros(sz);
    for pi = 1:size(gen_pos_new,1)
        pos(1) = gen_pos_new(pi,1);
        pos(2) = gen_pos_new(pi,2);
        RGB(pos(1),pos(2)) = gen_intensity_new(pi,1);
    end
    
    C = conv2(RGB, F, 'same');
    C = C/max(C(:))*255;
    Cim = 255-C;
    Cim(Cim>255) = 255;
    
    % Test image
    c_new = corr(IM(1:4:end)', Cim(1:4:end)');
    cvec = [cvec c_new];
    
    if c_new >= c_best
        c_best = c_new;
        gen_pos = gen_pos_new;
        gen_intensity = gen_intensity_new;

        figure(1)
        set(gcf,'position',[680 119 900 976])
        subplot(2,2,1)
        imagesc(Cim)
        colormap gray
        caxis([0 255])
        subplot(2,2,2)
        plot(cvec(end-100:end))
        ylabel('Correlation (fit)')
        pause(0.1)
        subplot(2,2,3)
        imagesc(IM)
        subplot(2,2,4)
        imagesc(double(IM/2)+Cim)
    end
    

end

cvec(end)

toc


%% Use Prior 


P = RGB;

F = fspecial('gaussian', round(size(P,1)/2), size(P,1)/60);

P = conv2(P, F, 'same');
close all
figure
imagesc(P/max(P(:)).*double(IM))
colormap gray

figure
imagesc(1-P)
colormap gray